//Copyright (c) 2015, NeoCortec A/S
//All rights reserved.
//
//Redistribution and use in source and binary forms, with or without
//modification, are permitted provided that the following conditions are met:
//
//1. Redistributions of source code must retain the above copyright notice,
//   this list of conditions and the following disclaimer.
//
//2. Redistributions in binary form must reproduce the above copyright notice,
//   this list of conditions and the following disclaimer in the documentation
//   and/or other materials provided with the distribution.
//
//3. Neither the name of the copyright holder nor the names of its contributors
//   may be used to endorse or promote products derived from this software
//   without specific prior written permission.
//
//THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
//AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
//IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
//DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
//FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
//DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
//SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
//CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
//OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
//OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

// This example configures one instance of NcApi.
// It shows a bare minimum implementation of what is required in order to receive and send messages.
// It is recommened to walk through ExampleSetupNcApi() and ExampleSendAcknowledged() defined at the end of this file.

#include "NcApi.h"
#include "NeoParser.h"

// Global declarations expected by NcApi
tNcApi g_ncApi[1];
uint8_t g_numberOfNcApis = 1;

// Global functions expected by NcApi
void NcApiTxData(uint8_t n, uint8_t * finalMsg, uint8_t finalMsgLength)
{
	// Write to particular UART indexed by n
}

void NcApiMessageWrittenCallback(uint8_t n, void * callbackToken, uint8_t * finalMsg, uint8_t finalMsgLength)
{
	
}


void NcApiMessageReceived(uint8_t n,void * callbackToken, uint8_t * msg, uint8_t msgLength)
{
	
tNcApiAllMessages NcApiMsg;

	switch(msg[0])
	{
	case HostAckEnum:
		NcApiGetMsgAsHostAck(msg, &NcApiMsg.NcApiHostAck);
		break;
	case HostNAckEnum:
		NcApiGetMsgAsHostAck(msg, &NcApiMsg.NcApiHostAck);
		break;
	case HostDataEnum:
		NcApiGetMsgAsHostData(msg, &NcApiMsg.NcApiHostData);
		break;
	case HostDataHapaEnum:
		NcApiGetMsgAsHostDataHapa(msg, &NcApiMsg.NcApiHostDataHapa);
		break;
	default:
		// unhandled 
		break;
	}
	
}
 

// Handle CTS and RX interrupts
void ExamplePassthroughCts()
{
	NcApiCallbackCtsActive(0);
}
void ExamplePassthroughRx(uint8_t byte)
{
	NcApiRxDataReceived(0, byte);
}

void ExamplePassthroughNwu()
{
	NcApiCallbackNwuActive(0);
}


// Sample implementation of how to send one message
void ExampleSendAcknowledged(uint16_t destNodeId, uint8_t port, uint8_t * payload, uint8_t payloadLen)
{
	tNcApiSendParams args;
	args.destNodeId = destNodeId;
	args.destPort = port;
	args.payload = payload;
	args.payloadLength = payloadLen;
	args.callbackToken = &g_ncApi;
	NcApiSendAcknowledged( 0, &args );
}

// Sample implementation of how to send one WesCmd
void ExampleSendWesCmd(NcApiWesCmdValues cmd)
{
	tNcApiWesCmdParams args;
	args.cmd = cmd;
	NcApiSendWesCmd( 0, &args );
}

// Sample implementation of how to send one WesResponse
void ExampleSendWesResponse(uint64_t uid, uint16_t nodeId)
{
	tNcApiWesResponseParams args;
	args.uid[0] = (uid >> 32) & 0xff;
	args.uid[1] = (uid >> 24) & 0xff;
	args.uid[2] = (uid >> 16) & 0xff;
	args.uid[3] = (uid >> 8) & 0xff;
	args.uid[4] = uid & 0xff;
	args.nodeId = nodeId;
	NcApiSendWesResponse( 0, &args );
}
// Sample implementation of how to setup global rx-handlers and initialize NcApi.
void ExampleSetupNcApi()
{
	NcApiInit();
	
}
