#include <errno.h>
#include <fcntl.h>
#include <poll.h>
#include <stdbool.h>
#include <stddef.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <sys/types.h>
#include "CtsTask.h"
#include "NcApi.h"
#include "NcPrintErr.h"
#include "Gpio.h"

const char * GpioUnExport(int pin)
{
	char buf[MAX_BUF];
	int fd = open("/sys/class/gpio/unexport", O_WRONLY);
	if (fd<0)
	{
		PrintErr("Failed to open: /sys/class/gpio/unexport. errno=%s", PrintErrNo(errno));
		return g_error;
	}
	sprintf(buf, "%d", pin);
	int len = strlen(buf);
	// dummy, the posix write is declared with __WUR
	ssize_t dummy= write(fd, buf, len);
	(void) dummy;
	close(fd);
	return NULL;
}

const char * GpioExport(int pin)
{
	char buf[MAX_BUF];
	int fd = open("/sys/class/gpio/export", O_WRONLY);
	if (fd<0)
	{
		PrintErr("Failed to open: /sys/class/gpio/export. errno=%s", PrintErrNo(errno));
		return g_error;
	}
	sprintf(buf, "%d", pin);
	int len = strlen(buf);
	if (write(fd, buf, len)!=len)
	{
		PrintErr("Failed to export GPIO pin %d. errno=%s", pin, PrintErrNo(errno));
		close(fd);
		return g_error;
	}
	close(fd);
	return NULL;
}

const char * GpioDirection(int pin, const char * dir)
{
	char buf[MAX_BUF];
	sprintf(buf, "/sys/class/gpio/gpio%d/direction", pin);
	int fd = open(buf, O_WRONLY);
	if (fd<0)
	{
		PrintErr("Failed to open: %s. errno=%s", buf, PrintErrNo(errno));
		return g_error;
	}
	int len = strlen(dir);
	if (write(fd, dir, len)!=len)
	{
		PrintErr("Failed to set direction on GPIO pin %d. errno=%s", pin, PrintErrNo(errno));
		close(fd);
		return g_error;
	}
	close(fd);
	return NULL;
}

const char * GpioTriggerMode(int pin, const char * mode)
{
	char buf[MAX_BUF];
	sprintf(buf, "/sys/class/gpio/gpio%d/edge", pin);
	int fd = open(buf, O_WRONLY);
	if (fd<0)
	{
		PrintErr("Failed to open: %s. errno=%s", buf, PrintErrNo(errno));
		return g_error;
	}
	int len = strlen(mode);
	if (write(fd, mode, len)!=len)
	{
		PrintErr("Failed to set trigger mode on GPIO pin %d to '%s'. errno=%s", pin, mode, PrintErrNo(errno));
		close(fd);
		return g_error;
	}
	close(fd);
	return NULL;
}

const char * GpioOpenValue(int pin, int * fd)
{
	char buf[MAX_BUF];
	sprintf(buf, "/sys/class/gpio/gpio%d/value", pin);
	*fd = open(buf, O_RDONLY);
	if (*fd>=0)
		return NULL;
	PrintErr("Failed to open: %s. errno=%s", buf, PrintErrNo(errno));
	return g_error;
}

