#include <stddef.h>
#include <stdint.h>
#include <stdlib.h>
#include <iostream>
#include "rapidjson/document.h"
#include "rapidjson/stringbuffer.h"
#include "rapidjson/writer.h"
#include "NeoParser.h"
#include "Serializer.h"
#include <langinfo.h>
#include <time.h>
#include <string.h>

using namespace rapidjson;
using namespace std;

static const uint8_t Nibbles[16] = {'0','1','2','3','4','5','6','7','8','9','a','b','c','d','e','f'};

typedef class tSerializerJson : public ISerializer
{
	const char * ToHexString(uint8_t * ba, uint8_t len)
	{
		char* dst = (char*)malloc(2*len+1);
		int i;
		for (i=0; i<len; i++)
		{
			uint8_t b = ba[i];
			dst[i*2+0] = Nibbles[ b >> 4 ];
			dst[i*2+1] = Nibbles[ b & 0xf ];
		}
		dst[ len << 1 ] = 0;
		return dst;
	}

	void WritePayload(Writer<StringBuffer>& w, uint8_t * ba, uint8_t len)
	{
		const char * p = ToHexString(ba, len);
		w.String("payloadHex");
		w.String(p);
		free((void*)p);
	}

	void WritePayloadRaw(Writer<StringBuffer>& w, uint8_t * ba, uint8_t len)
	{
		w.String("payload");
		w.StartArray();
		for (unsigned i = 0; i < len; i++)
		    w.Uint(ba[i]);
		w.EndArray();
	}


	void WriteMsgType(Writer<StringBuffer>& w, NcApiMessageType mt)
	{

		w.String("objectType");
		switch(mt)
		{
		case CommandAcknowledgedEnum:
			w.String("sendPayload");
			break;
	
		case HostAckEnum:
			w.String("ack");
			break;
		case HostNAckEnum:
			w.String("nack");
			break;
			
		case WesCmdEnum:
			w.String("wesCmd");
			break;

		case WesStatusEnum:
			w.String("wesStatus");
			break;
			
		case WesRequestEnum:
			w.String("wesRequest");
			break;
		case WesResponseEnum:
			w.String("wesResponse");
			break;
			
		case HostDataEnum:
		case HostDataHapaEnum:
			w.String("receivedPayload");
			break;
	
		
	
		default:
			w.String("unknown");
			break;
		}
		
	}
	void WriteGwTimestamp(Writer<StringBuffer>& w)
	{
	struct timespec ts;
 
	char datestring[32];
	char datestringEx[32];
	long ms; // Milliseconds
	
		timespec_get(&ts, TIME_UTC);
		ms = (ts.tv_nsec / 1.0e6); // Convert nanoseconds to milliseconds
	
	
		strftime(datestring, sizeof datestring, "%FT%T", gmtime(&ts.tv_sec));
		sprintf(datestringEx,"%s.%03ldZ",datestring,ms % 1000);
		w.String("gwTimestamp");
		w.String(datestringEx);

		
	}

	void WriteOriginId(Writer<StringBuffer>& w, uint16_t v)
	{
		w.String("nodeId");
		w.Uint(v);
	}

	void WriteUid(Writer<StringBuffer>& w, uint8_t v[5])
	{
		const char * p = ToHexString(v, 5);
		w.String("uidHex");
		w.String(p);
		free((void*)p);
	}

	void WritePort(Writer<StringBuffer>& w, uint8_t v)
	{
		w.String("port");
		w.Uint(v);
	}

	void WritePackageAge(Writer<StringBuffer>& w, uint16_t v)
	{
		w.String("packageAgeMicroSecs");
		w.Uint64((uint64_t(v)) * 125000UL);
	}

	void WritePackageAgeHapa(Writer<StringBuffer>& w, uint32_t v)
	{
		w.String("packageAgeMicroSecs");
		w.Uint64((uint64_t(v)) * 1000000UL / 524288UL);
	}

	void WriteStatus(Writer<StringBuffer>& w, const char * s)
	{
		w.String("status");
		w.String(s);
	}

	void WriteContent(Writer<StringBuffer>& w, const char * s)
	{
		w.String("content");
		w.String(s);
	}

	void WriteWesStatus(Writer<StringBuffer>& w, NcApiWesStatusValues v)
	{
		w.String("status");
		w.Uint(v);
	}

	void WriteCfngValAppFunction(Writer<StringBuffer>& w, uint8_t v)
	{
		w.String("AppFunction");
		w.Uint(v);
	}

	const char * ParseJson(const char * json, int jsonLen, Document * d)
	{
		const char * m = strndup(json, jsonLen);

		if (*m!='{')
		{
			free((void*)m);
			return "Message is not JSON-formatted";
			
		}
		if (d->Parse(m).HasParseError())
		{
			free((void*)m);
			return "Message is not JSON-formatted";
		}
		free((void*)m);
		return NULL;
	}

	const char * RaiseMissingMember(const char * member)
	{
		char * p = (char*)malloc(1024);
		snprintf(p, 1024, "Missing member: %s", member);
		return p;
	}

	const char * ReadUInt8(Document& d, const char * member, uint8_t * p, bool raiseOnError = true)
	{
		if (!d.HasMember(member))
		{
			if (raiseOnError)
				return RaiseMissingMember(member);
			return NULL;
		}
		*p = d[member].GetUint();
		return NULL;
	}

	const char * ReadUInt16(Document& d, const char * member, uint16_t * p, bool raiseOnError = true)
	{
		if (!d.HasMember(member))
		{
			if (raiseOnError)
				return RaiseMissingMember(member);
			return NULL;
		}
		*p = d[member].GetUint();
		return NULL;
	}

	const char * ReadUInt32(Document& d, const char * member, int * p, bool raiseOnError = true)
	{
		if (!d.HasMember(member))
		{
			if (raiseOnError)
				return RaiseMissingMember(member);
			return NULL;
		}
		*p = d[member].GetUint();
		return NULL;
	}

	uint8_t HexCharToNibble(char c)
	{
		if (c<='9') return c - '0';
		if (c<='F') return c - 'A' + 10;
		return c - 'a' + 10;
	}

	const char * ReadHexString(Document& d, const char * member, uint8_t * payload, uint8_t len, bool raiseOnError = true)
	{
		if (!d.HasMember(member))
		{
			if (raiseOnError)
				return RaiseMissingMember(member);
			return NULL;
		}
		const char * hex = d[member].GetString();
		int hexlen = strlen(hex);
		if ((hexlen >> 1) != len)
			return strdup("HexString length mismatch");
		int i;
		for (i=0; i<len; i++)
			payload[i] = (HexCharToNibble(hex[2*i+0]) << 4)
					   | (HexCharToNibble(hex[2*i+1]));
		return NULL;
	}

	const char * ReadString(Document& d, const char * member)
	{
		if (!d.HasMember(member))
		{
			return NULL;
		}
		return d[member].GetString();
	}

	const char * ReadIntArray(Document& d, const char * member, uint8_t ** payload, uint8_t * len, bool raiseOnError = true)
	{
		if (!d.HasMember(member))
		{
			if (raiseOnError)
				return RaiseMissingMember(member);
			return NULL;
		}
		if (!d[member].IsArray())
		{
			if (raiseOnError)
				return RaiseMissingMember(member);
			return NULL;
		}
		Value& p = d[member];
		*len = p.Size();
		if (*len == 0)
		{
			*payload = NULL;
			return NULL;
		}
		*payload = (uint8_t*)malloc(*len);
		for (SizeType i = 0; i < p.Size(); i++) // Uses SizeType instead of size_t
		        (*payload)[i]=p[i].GetInt();

		return NULL;
	}

public:
	virtual ~tSerializerJson() {}
	virtual const char * SerializeAck(tNcApiHostAck * p)
	{
		StringBuffer s;
		Writer<StringBuffer> w(s);
		w.StartObject();
		WriteMsgType(w, HostAckEnum);
		WriteGwTimestamp(w);
		WriteOriginId(w, p->originId);
		w.EndObject();
		return strdup(s.GetString());
	}

	virtual const char * SerializeNAck(tNcApiHostAck * p)
	{
		StringBuffer s;
		Writer<StringBuffer> w(s);
		w.StartObject();
		WriteMsgType(w, HostNAckEnum);
		WriteGwTimestamp(w);
		WriteOriginId(w, p->originId);
		w.EndObject();
		return strdup(s.GetString());
	}

	virtual const char * Serialize(tNcApiHostData * p)
	{
		StringBuffer s;
		Writer<StringBuffer> w(s);
		w.StartObject();
		WriteMsgType(w, HostDataEnum);
		WriteGwTimestamp(w);
		WriteOriginId(w, p->originId);
		WritePort(w, p->port);
		WritePackageAge(w, p->packageAge);
		w.String("packageAgeType");
		w.String("normal");		
    	WritePayloadRaw(w, p->payload, p->payloadLength);
		w.EndObject();
		return strdup(s.GetString());
	}

	virtual const char * SerializeRaw(uint8_t * Buffer,uint8_t payloadLength)
	{
		StringBuffer s;
		Writer<StringBuffer> w(s);
		w.StartObject();
		WriteMsgType(w, (NcApiMessageType)Buffer[0]);
		WriteGwTimestamp(w);
    	WritePayloadRaw(w, &Buffer[0], payloadLength);
		w.EndObject();
		return strdup(s.GetString());
	}

	virtual const char * Serialize(tNcApiHostDataHapa * p)
	{
		StringBuffer s;
		Writer<StringBuffer> w(s);
		w.StartObject();
		WriteMsgType(w, HostDataHapaEnum);
		WriteGwTimestamp(w);
		WriteOriginId(w, p->originId);
		WritePort(w, p->port);
		WritePackageAgeHapa(w, p->packageAge);

		w.String("packageAgeType");
		w.String("hapa");		

		//WritePayload(w, p->payload, p->payloadLength);
		WritePayloadRaw(w, p->payload, p->payloadLength);
		w.EndObject();
		return strdup(s.GetString());
	}

	virtual const char * Serialize(tNcApiWesRequest * p)
	{
		StringBuffer s;
		Writer<StringBuffer> w(s);
		w.StartObject();
		WriteMsgType(w, WesRequestEnum);
		WriteGwTimestamp(w);
		WriteUid(w, p->uid);
		WriteCfngValAppFunction(w, p->cfngValApplication);
		w.EndObject();
		return strdup(s.GetString());
	}

	virtual const char * Serialize(tNcApiWesStatus * p)
	{
		StringBuffer s;
		Writer<StringBuffer> w(s);
		w.StartObject();
		WriteMsgType(w, WesStatusEnum);
		WriteGwTimestamp(w);
		WriteWesStatus(w, (NcApiWesStatusValues)p->Status);
		w.EndObject();
		return strdup(s.GetString());
	}

	virtual const char * Serialize(ReplyMsg * p)
	{
		StringBuffer s;
		Writer<StringBuffer> w(s);
		w.StartObject();
		
		w.String("objectType");
		w.String("replyMsg");
				
		WriteGwTimestamp(w);
		
		
		WriteStatus(w, p->status);
		if (p->content==NULL)
			WriteContent(w, "");
		else
			WriteContent(w, p->content);
		
		w.EndObject();
		return strdup(s.GetString());
	}

	virtual const char * DeSerialize(const char * m, int len, SendMsg * msg)
	{
		Document d;
		
		const char * ret = ParseJson(m, len, &d);
		const char * str; 
	
		if (ret==NULL)
		{
			str=ReadString(d,"objectType");
			
			if (str!=NULL)
			{
				
				if (strcmp(str,"sendPayload")==0)
				{
					
					
					msg->msgType = (NcApiMessageType)CommandAcknowledgedEnum;
					
					ret = ReadIntArray(d, "payload", &msg->payload, &msg->payloadLength);
					if (ret==NULL) ret = ReadUInt16(d, "nodeId", &msg->destNodeId);
					if (ret==NULL) ret = ReadUInt8(d, "port", &msg->destPort);

				}
				else
				{
					ret="";
				}
			}
			else
			{
				ret="";
			}
		}
	
		return ret;
	}

	virtual const char * DeSerialize(const char * m, int len, tNcApiWesCmd * msg)
	{
		Document d;
		const char * ret = ParseJson(m, len, &d);
		uint8_t ncApiWesCmd = 0;
		const char * str; 

		if (ret==NULL)
		{
			str=ReadString(d,"objectType");
			if (str!=NULL)
			{
				
				if (strcmp(str,"wesCmd")==0)
				{
					
					ret = ReadUInt8(d, "cmd", &ncApiWesCmd);
					if (ret==NULL) msg->cmd = (NcApiWesCmdValues)ncApiWesCmd;

				}
				else
				{
					ret="";
				}
			}
			else
			{
				ret="";
			}
		}
	

		return ret;
	}

	virtual const char * DeSerialize(const char * m, int len, tNcApiAltCmd * msg)
	{
		Document d;
		const char * ret = ParseJson(m, len, &d);
		uint8_t ncApiAltCmd = 0;
		ret="";
		
		return ret;
	}

	virtual const char * DeSerialize(const char * m, int len, tNcApiWesResponse * msg)
	{
		Document d;
		const char * ret = ParseJson(m, len, &d);
		const char * str; 
		

		if (ret==NULL)
		{
			str=ReadString(d,"objectType");
			if (str!=NULL)
			{
				
				if (strcmp(str,"wesResponse")==0)
				{
					
					ret = ReadUInt16(d, "nodeId", &msg->nodeId);
					if (ret==NULL)
					{
						uint8_t uidLen = (uint8_t)sizeof(msg->uid);
						ret = ReadHexString(d, "uidHex", msg->uid, uidLen);
					}
					
				}
				else
				{
					ret="";
				}
			}
			else
			{
				ret="";
			}
		}

		return ret;
	}
} SerializerJson;
